# Authors: Hugo Richard, Pierre Ablin
# License: BSD 3 clause

from ._multiviewica import multiviewica, _multiview_ica_main
from ._groupica import groupica
from ._permica import permica, _hungarian

__version__ = '0.1'
